import sys
import os
from pathlib import Path
from typing import Any, Tuple


# Add the src directory to the Python path
sys.path.insert(
    0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "src"))
)

from entoli.base.maybe import Just
from entoli.system import create_dir_if_missing, file_exists, write_file
from entoli.process import (
    CreateProcess,
    create_process,
    h_get_contents,
    wait_for_process,
)
from entoli.prelude import unlines, put_strln
from entoli.base.io import Io

flake_content = """
{
  description = "A Python 3.12 development environment";

  inputs = {
    nixpkgs.url = "github:NixOS/nixpkgs/nixos-unstable"; # Switched to unstable for more recent packages
    flake-utils.url = "github:numtide/flake-utils";
  };

  outputs = { self, nixpkgs, flake-utils, ... }:
    flake-utils.lib.eachDefaultSystem (system:
      let
        pkgs = import nixpkgs {
          inherit system;
          config = {
            allowUnfree = true; # If necessary for any unfree packages
          };
        };

        pythonEnv = pkgs.python312.withPackages (ps: with ps; [
          ps.pip
        ]);

      in
      {
        devShell = pkgs.mkShell {
          buildInputs = [ pythonEnv ];
          shellHook = ''
            if [ ! -d "venv" ]; then
              echo "Creating a Python virtual environment..."
              ${pythonEnv}/bin/python -m venv venv
            fi

            echo "Activating Python virtual environment..."
            source venv/bin/activate

            if [ ! -f "requirements_dev.txt" ]; then
              echo "Creating a requirements_dev.txt file..."
              touch requirements_dev.txt
            fi
            pip install -r requirements_dev.txt
          '';
        };
      }
    );
}

"""
flake_path = Path(__file__).parent.absolute() / "build"


# Call nix d
def flake_call(flake_path: Path, cmd: str) -> Io[Tuple[int, str, str]]:
    proc_spec = CreateProcess(
        command=f'nix develop --command bash -c "{cmd}"',
        use_shell=True,
        cwd=Just(flake_path),
    )

    def _inner(tpl) -> Io[Tuple[int, str, str]]:
        match tpl:
            case Just(stdin), Just(stdout), Just(stderr), process:
                return wait_for_process(process).and_then(
                    lambda exit_code: h_get_contents(stdout).and_then(
                        lambda out: h_get_contents(stderr).and_then(
                            lambda err: Io.pure((exit_code, out, err))
                        )
                    )
                )
            case _:
                raise RuntimeError("Failed to create process")

    return create_process(proc_spec).and_then(_inner)


main = (
    create_dir_if_missing(True, flake_path)
    .then(write_file(flake_path / "flake.nix", flake_content))
    .then(file_exists(flake_path / "flake.nix"))
    .and_then(lambda res: put_strln(f"Flake file exists: {res}"))
    .then(write_file(flake_path / ".envrc", "use flake"))
    .then(file_exists(flake_path / ".envrc"))
    .and_then(lambda res: put_strln(f"Envrc file exists: {res}"))
    .then(flake_call(flake_path, "pwd && which python"))
    .then(
        flake_call(
            flake_path,
            unlines(
                [
                    "pip install django",
                    "django-admin --version",
                    "django-admin startproject auto_project",
                    "cd auto_project",
                    "python manage.py startapp auto_app_0",
                    "python manage.py startapp auto_app_1",
                    "python manage.py makemigrations",
                    "python manage.py migrate",
                    "export DJANGO_SUPERUSER_USERNAME=admin",
                    "export DJANGO_SUPERUSER_EMAIL=a@a.com",
                    "export DJANGO_SUPERUSER_PASSWORD=admin",
                    "python manage.py createsuperuser --noinput",
                ]
            ),
        )
    )
    .and_then(
        lambda res: put_strln(f"exit code: {res[0]}")
        .then(put_strln(f"stdout: {res[1]}"))
        .then(put_strln(f"stderr: {res[2]}"))
    )
)

if __name__ == "__main__":
    main.action()



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Generic, Protocol, TypeVar

from .typeclass import Monad

_A = TypeVar("_A")
_B = TypeVar("_B")


# class _Ei(Monad[_A], Protocol[_A]):
#     @staticmethod
#     def fmap(f: Callable[[_A], _B], x: _Ei[_A]) -> _Ei[_B]: ...

#     @staticmethod
#     def pure(x: _A) -> _Ei[_A]: ...

#     @staticmethod
#     def ap(f: _Ei[Callable[[_A], _B]], x: _Ei[_A]) -> _Ei[_B]: ...

#     @staticmethod
#     def bind(x: _Ei[_A], f: Callable[[_A], _Ei[_B]]) -> _Ei[_B]: ...

#     def map(self, f: Callable[[_A], _B]) -> _Ei[_B]: ...

#     def and_then(self, f: Callable[[_A], _Ei[_B]]) -> _Ei[_B]: ...

#     def then(self, x: _Ei[_B]) -> _Ei[_B]: ...

#     def unwrap(self) -> _A: ...

#     def unwrap_or(self, default: _A) -> _A: ...


@dataclass
class Left(Generic[_A]):
    value: _A

    def __repr__(self) -> str:
        return f"Left({self.value})"

    def __str__(self) -> str:
        return repr(self)

    def __bool__(self) -> bool:
        return False

    @staticmethod
    def fmap(f: Callable[[_A], _B], x: Left) -> Ei[_B]:
        return Left()

    @staticmethod
    def pure(x: _A) -> Ei[_A]:
        return Left()

    @staticmethod
    def ap(f: Left, x: Left) -> Ei[Any]:
        return Left()

    @staticmethod
    def bind(x: Left, f: Callable[[_A], Left]) -> Ei[_A]:
        return Left()

    def map(self, f: Callable[[_A], _B]) -> Ei[_B]:
        return Left()

    def and_then(self, f: Callable[[_A], Ei[_B]]) -> Left:
        return Left()

    def then(self, x: Left) -> Ei[Any]:
        return Left()

    def unwrap(self) -> Any:
        raise ValueError("Nothing.unwrap: cannot unwrap Nothing")

    def unwrap_or(self, default: _A) -> _A:
        return default


@dataclass
# class Right(_Ei[_A]):
class Right(Generic[_A]):
    value: _A

    def __repr__(self) -> str:
        return f"Right({self.value})"

    def __str__(self) -> str:
        return self.__repr__()

    def __bool__(self) -> bool:
        return True

    @staticmethod
    def fmap(f: Callable[[_A], _B], x: Right[_A]) -> Ei[_B]:
        return Right(f(x.value))

    @staticmethod
    def pure(x: _A) -> Ei[_A]:
        return Right(x)

    @staticmethod
    def ap(f: Right[Callable[[_A], _B]], x: Right[_A]) -> Ei[_B]:
        return Right(f.value(x.value))

    @staticmethod
    def bind(x: Right[_A], f: Callable[[_A], Right[_B]]) -> Ei[_B]:
        return f(x.value)

    def map(self, f: Callable[[_A], _B]) -> Ei[_B]:
        return Right(f(self.value))

    def and_then(self, f: Callable[[_A], Right[_B]]) -> Ei[_B]:
        return f(self.value)

    def then(self, x: Right[_B]) -> Ei[_B]:
        return x

    def unwrap(self) -> _A:
        return self.value

    def unwrap_or(self, default: _A) -> _A:
        return self.value


Either = Left[_A] | Right[_B]



from datetime import timedelta, datetime
from time import sleep

from entoli.base.io import Io


def loop(io: Io[None]) -> Io[None]:
    def _inner() -> None:
        while True:
            io.action()

    return Io(_inner)


def delay_for(delay: timedelta) -> Io[None]:
    return Io(lambda: sleep(delay.total_seconds()))


def delay_until(time_point: datetime) -> Io[None]:
    return Io(lambda: sleep((time_point - datetime.now()).total_seconds()))



from setuptools import setup, find_packages

setup(
    name="entoli",
    version="0.1.1",
    author="Your Name",
    author_email="cwahn0904@gmail.com",
    description="A Python functional programming library",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/cwahn/entoli",  # Update this URL to your project's repository
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.12",
    install_requires=[],
    extras_require={
        "dev": open("requirements_dev.txt").read().splitlines(),
    },
    entry_points={
        "console_scripts": [
            # Add any CLI scripts here
        ],
    },
)
